unit TabListBox;
{----------------------------------------------------------------------------}
{  TListBoxRk component                                                      }
{  ====================                                                      }
{                                                                            }
{  Properties - public                                                       }
{  -------------------                                                       }
{  class function PixelToMm(Value:Integer):integer;                          }
{  class function MmToPixel(Value:integer):integer;                          }
{                                                                            }
{  procedure SetAllTabTo(Value:integer=100; ValueA:TAlignment=taLeftJustify);}
{  - set distance bettwen tabs to Value and Alignment to ValueA              }
{  Example: SetAllTabTo(50);                                                 }
{  Result: Tab[1]=50, Tab[2]=100, Tab[3]=150 .. Tab[TabCount]=Value*TabCount }
{                                                                            }
{  property TabAlign[Index:byte]:TAlignment;                                 }
{  - get or set alignment of tab                                             }
{  Example: TabAlign[1]:=taRightJustify                                      }
{                                                                            }
{  property Tab[Index:byte]:Integer;                                         }
{  - get or set position of tab from left edge TListBox in pixels            }
{  - values of tabs should be in ascending order                             }
{  - Index range is 1 to TabCount                                            }
{  Example: Tab[1]:=100; Tab[2]:=150; Tab[3]:=400                            }
{                                                                            }
{                                                                            }
{  Properties - published                                                    }
{  ----------------------                                                    }
{  property HintText:Boolean;                                                }
{  - when true then automatic show hint window for items which not fit       }
{    to ListBox width                                                        }
{                                                                            }
{  property OnMouseEnter:TNotifyEvent;                                       }
{  - occurs when the mouse pointer moves to control                          }
{                                                                            }
{  property OnMouseExit:TNotifyEvent;                                        }
{  - occurs when the mouse pointer moves from control                        }
{                                                                            }
{  property TabCount:byte;                                                   }
{  - get or set count of tab                                                 }
{  - after create TabCount equal to 9                                        }
{                                                                            }
{  property TabString:string;                                                }
{  - get or set values of positions and alignments for all tabs              }
{  - format is: (pos[|alg]);                                                 }
{  - pos is tab position in pixels                                           }
{  - alg is alignment of tab where:                                          }
{    0 - taLeftJustify                                                       }
{    1 - taRightJustify                                                      }
{    2 - taCenter                                                            }
{  - parameter alg is optionally                                             }
{  Example: TabString:='100;150|0;220|2;300';                                }
{                                                                            }
{                                                                            }
{  Removed properties from standard TListBox                                 }
{  -----------------------------------------                                 }
{  Style                                                                     }
{  TabWidth                                                                  }
{  OnDrawItem                                                                }
{  OnMeasureItem                                                             }
{                                                                            }
{                                                                            }
{  History                                                                   }
{  -------                                                                   }
{                                                                            }
{  Version 1.1 (20.10.2001)                                                  }
{  Fixed: setting HintText to True from Object Inspector have no effect      }
{  Fixed: bad drawing items, selected items not show                         }
{  Added: descendant from TAction for work with clipboard                    }
{                                                                            }
{  Version 1.0 (9.10.2001)                                                   }
{  -initial version                                                          }
{                                                                            }
{                                                                            }
{  License                                                                   }
{  -------                                                                   }
{  It is freeware.                                                           }
{                                                                            }
{                                                                            }
{  Contact information                                                       }
{  -------------------                                                       }
{  Radek Kouba                                                               }
{  r.kouba@tkrostrov.cz                                                      }
{                                                                            }
{----------------------------------------------------------------------------}

interface

uses
 SysUtils, Windows, Graphics, Classes, Messages, Controls, StdCtrls;

type
  TListBoxRk=class;

  THintWindowRk = class(THintWindow)
  private
   Lb:TListBoxRk;
  public
   procedure Paint;override;
  end;

  TListBoxRk=class(TCustomListBox)
  private
    FTabCount:byte;
    FTabs:array of Integer;
    FTabAlign:array of TAlignment;
    FHintText:Boolean;
    FHintWin:THintWindowRk;
    FOnMouseExit: TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    procedure SetHintText(Value:boolean);
    function GetTabStops(Index:byte):Integer;
    procedure SetTabStops(Index:byte;Value:Integer);
    function GetTabAlign(Index:byte):TAlignment;
    procedure SetTabAlign(Index:byte;Value:TAlignment);
    function GetTabString: string;
    procedure SetTabString(Value: string);
    procedure SetTabCount(Value: byte);
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  protected
    procedure LbDrawItem(Control: TWinControl;
      Index: Integer; Rect: TRect; State: TOwnerDrawState);
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class function PixelToMm(Value:Integer):integer;
    class function MmToPixel(Value:integer):integer;
    procedure DefaultHandler(var Message); override;
    procedure SetAllTabTo(Value:integer=100; ValueA:TAlignment=taLeftJustify);
    property TabAlign[Index:byte]:TAlignment read GetTabAlign write SetTabAlign;
    property Tab[Index:byte]:Integer read GetTabStops write SetTabStops;
  published
    property Align;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property Color;
    property Columns;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property HintText:Boolean read FHintText write SetHintText default true;
    property ImeMode;
    property ImeName;
    property IntegralHeight;
    property ItemHeight;
    property Items;
    property MultiSelect;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseEnter:TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseExit:TNotifyEvent read FOnMouseExit write FOnMouseExit;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
    property TabCount:byte read FTabCount write SetTabCount;
    property TabString:string read GetTabString write SetTabString;
  end;

implementation

const
 TabIndexError='Index out of bounds (%d)';

procedure DrawItemText(const Rect:TRect; const Canvas:TCanvas;
  const Lb:TListBoxRk; s:string);
var
 s1:string;
 i,CurTab,x,LastPos:Integer;
begin
 s:=TrimRight(s);
 if pos(#9,s)=0 then
 begin
  Canvas.TextRect(Rect,Rect.Left+2,Rect.Top,s);
 end
 else
 begin
  s:=s+#9;
  CurTab:=0;
  repeat
   i:=pos(#9,s);
   if i=1 then LastPos:=Rect.Left+Lb.FTabs[CurTab]
   else
   begin
    s1:=Copy(s,1,i-1);
    case Lb.FTabAlign[CurTab] of // must Rect.Left+ when ListBox.Columns>1
     taLeftJustify:x:=Rect.Left+Lb.FTabs[CurTab]+2;
     taRightJustify:x:=Rect.Left+Lb.FTabs[CurTab]-Canvas.TextWidth(s1);
     taCenter:x:=Rect.Left+Lb.FTabs[CurTab]-(Canvas.TextWidth(s1) div 2);
    end;
    Canvas.TextRect(Rect,x,Rect.Top,s1);
    LastPos:=x+Canvas.TextWidth(s1);
   end;
   s:=Copy(s,i+1,Length(s));
   Inc(CurTab);
   // check validity of tabs
   if(s<>'')and(CurTab<Lb.FTabCount+1)and(Lb.FTabs[CurTab-1]>=Lb.FTabs[CurTab])then
   begin
    SetLength(s,Length(s)-1);
    while pos(#9,s)>0 do Delete(s,pos(#9,s),1);
    Canvas.TextRect(Rect,LastPos,Rect.Top,s);
    Exit;
   end;
  until(s='')or(CurTab=Lb.FTabCount+1);
 end;
end;

// THintWindowRk
procedure THintWindowRk.Paint;
var
 Rect:TRect;
begin
 Rect:=ClientRect;
 Inc(Rect.Top);
 DrawItemText(Rect,Canvas,Lb,Caption);
end;

// TListBoxRk

class function TListBoxRk.MmToPixel(Value:Integer):Integer;
begin
 Result:=Round(Value/0.376);
end;

class function TListBoxRk.PixelToMm(Value:Integer):Integer;
begin
 Result:=Round(Value*0.376);
end;

constructor TListBoxRk.Create(AOwner: TComponent);
var
 i:Integer;
begin
 inherited;
 FTabCount:=9;
 SetLength(FTabs,FTabCount+1);
 SetLength(FTabAlign,FTabCount+1);
 for i:=1 to FTabCount do FTabs[i]:=i*100;
 Style:=lbOwnerDrawFixed;
 OnDrawItem:=LbDrawItem;
 FHintWin:=THintWindowRk.Create(self);
 FHintwin.Color:=clInfoBk;
 FHintwin.Canvas.Font.Color:=clInfoText;
 FHintWin.Lb:=self;
 FHintText:=True;
end;

destructor TListBoxRk.Destroy;
begin
 FTabs:=nil;
 FTabAlign:=nil;
 inherited;
end;

procedure TListBoxRk.SetHintText(Value:boolean);
begin
 FHintText:=Value;
 if not FHintText then FHintWin.ReleaseHandle;
end;

function TListBoxRk.GetTabStops(Index:byte):Integer;
begin
 if Index in [0..FTabCount] then Result:=FTabs[Index]
 else
  raise Exception.CreateFmt(TabIndexError,[Index]);
end;

procedure TListBoxRk.SetTabStops(Index:byte;Value:Integer);
begin
 if(Index in [1..FTabCount])and(FTabs[Index]<>Value)then
 begin
  FTabs[Index]:=Value;
  Invalidate;
 end;
end;

function TListBoxRk.GetTabAlign(Index:byte):TAlignment;
begin
 if Index in [0..FTabCount] then Result:=FTabAlign[Index]
 else
  raise Exception.CreateFmt(TabIndexError,[Index]);
end;

procedure TListBoxRk.SetTabAlign(Index:byte;Value:TAlignment);
begin
 if(Index in [1..FTabCount])and(FTabAlign[Index]<>Value)then
 begin
  FTabAlign[Index]:=Value;
  Invalidate;
 end;
end;

function TListBoxRk.GetTabString: string;
var
 i:Integer;
begin
 result:='';
 for i:=1 to FTabCount do
  Result:=Result+IntToStr(FTabs[i])+'|'+IntToStr(Ord(FTabAlign[i]))+';';
 SetLength(Result,Length(Result)-1);
end;

procedure TListBoxRk.SetTabString(Value: string);
var
 s:string;
 i,CurTab:Integer;
 pInt:PInteger;
begin
 if Value='' then Exit;
 pInt:=PInteger(FTabs);
 FillChar(pInt^,SizeOf(integer)*(FTabCount+1),0);
 pInt:=PInteger(FTabAlign);
 FillChar(pInt^,SizeOf(TAlignment)*(FTabCount+1),taLeftJustify);
 if Value[length(value)]<>';' then value:=value+';';
 CurTab:=1;
 repeat
  i:=pos(';',value);
  s:=Copy(value,1,i-1);
  if Pos('|',s)=0 then
  begin
   FTabs[CurTab]:=StrToIntDef(s,0);
  end
  else
  begin
   FTabs[CurTab]:=StrToIntDef(Copy(s,1,Pos('|',s)-1),0);
   case StrToIntDef(Copy(s,Pos('|',s)+1,1),0) of
    // 0:; is on FillChar
    1:FTabAlign[CurTab]:=taRightJustify;
    2:FTabAlign[CurTab]:=taCenter;
   end;
  end;
  value:=Copy(value,i+1,Length(value));
  Inc(CurTab);
 until(Value='')or(CurTab=FTabCount+1);
 Invalidate;
end;

procedure TListBoxRk.SetTabCount(Value: byte);
var
 i:integer;
begin
 if(Value>0)and(FTabCount<>Value)then
 begin
  SetLength(FTabs,Value+1);
  SetLength(FTabAlign,Value+1);
  // set new tabs to zero
  for i:=FTabCount+1 to Value do
  begin
   FTabs[i]:=0;
   FTabAlign[i]:=taLeftJustify;
  end;
  FTabCount:=Value;
 end;
 Invalidate;
end;

procedure TListBoxRk.LbDrawItem(Control: TWinControl;
 Index: Integer; Rect: TRect; State: TOwnerDrawState);
begin
 Canvas.FillRect(Rect);
 Canvas.Brush.Style:=bsClear;
 DrawItemText(Rect,Canvas,self,Items[Index]);
end;

// procedury a funkce pro Hint text

procedure TListBoxRk.CMMouseEnter(var Message: TMessage);
begin
 inherited;
 if assigned(FOnMouseEnter) then FOnMouseEnter(self);
end;

procedure TListBoxRk.CMMouseLeave(var Message: TMessage);
begin
 inherited;
 FHintWin.ReleaseHandle;
 If assigned(FOnMouseExit) then FOnMouseExit(self);
end;

procedure TListBoxRk.DefaultHandler(var Message);
var
 msgT:TMsg;
begin
 inherited;
 if(FHintWin=nil)or not(FHintText)then Exit;
 // convert message to TMsg
 MsgT.message:=TMessage(Message).Msg;
 if(TMessage(Message).Msg = WM_VScroll)or
   (FHintWin.IsHintMsg(MsgT) and
    //if it's WM_Mousemove then we don't want to hide the hint window
   (MsgT.message <> WM_MouseMove))
 then FHintWin.ReleaseHandle;
end;

procedure TListBoxRk.MouseMove(Shift: TShiftState; X, Y: Integer);
 function WidthOfItem(s:string):Integer;
 var
  i,PocTab,CurTab:integer;
 begin
  s:=TrimRight(s);
  if(pos(#9,s)=0)then
   Result:=Canvas.textwidth(s)
  else
  begin
   // determine count of tabs in string
   PocTab:=0;
   for i:=1 to Length(s)do
    if s[i]=#9 then Inc(PocTab);
   // check validity of tabs
   CurTab:=PocTab;
   for i:=2 to PocTab do
    if FTabs[i-1]>=FTabs[i]then
    begin
     CurTab:=i-1;
     Break;
    end;
   // search position of CurTab
   PocTab:=0;
   for i:=1 to Length(s)do
   begin
    if s[i]=#9 then Inc(PocTab);
    if CurTab=PocTab then Break;
   end;
   s:=Copy(s,i+1,Length(s));
   if CurTab=PocTab then
   begin
    case FTabAlign[CurTab] of
     taLeftJustify:
      Result:=FTabs[CurTab]+2+Canvas.textwidth(s);
     taRightJustify:
      Result:=FTabs[CurTab]-2;
     taCenter:
      Result:=FTabs[CurTab]+(Canvas.TextWidth(s) div 2)-2;
    end;
   end
   else
   begin
    while pos(#9,s)>0 do Delete(s,pos(#9,s),1);
    Result:=FTabs[CurTab]+2+Canvas.textwidth(s);
   end;
  end;
 end;
var
 sirka,ind:integer;
 rct:TRect;
begin
 inherited;
 if not FHintText then exit;
 // Get the index of the item under the mouse pointer
 ind:=ItemAtPos(Point(x,y),true);
 // Check if it is a legal item
 if ind = -1 then
 begin
  FHintWin.ReleaseHandle;
  exit;
 end;
 // Get the default item coordinates
 rct:=ItemRect(ind);
 // Check if the Item Text is wider then the cliprectangle
 sirka:=WidthOfItem(Items[Ind]);
 if sirka>(rct.Right-rct.Left-3)then
 begin
  // Stretch it to fit the whole item text
  Rct.Right:=Rct.Left+sirka+7;
  Dec(Rct.Bottom,3);
  OffsetRect(rct,-1,-1);
  // now convert to screen coordinates so that THintWindow can use them
  rct.TopLeft:=ClientToScreen(rct.TopLeft);
  rct.BottomRight:=ClientToScreen(rct.BottomRight);
  // and show
  FHintWin.ActivateHint(rct,items[ind]);
 end
 else FHintWin.ReleaseHandle;
end;

procedure TListBoxRk.SetAllTabTo(Value:integer=100; ValueA:TAlignment=taLeftJustify);
var
 i:integer;
begin
 if Value>0 then
 begin
  for i:=1 to FTabCount do
  begin
   FTabs[i]:=Value;
   FTabAlign[i]:=ValueA;
  end;
  Invalidate;
 end;
end;

end.
